import abc
import json
import logging
from typing import Optional, Awaitable

from tornado import web

from .param_check import sign_check
from ...logger import LOG

_RESP_BAD_REQUEST = {'code': '5101', 'message': ['Bad request: fail to parse body as JSON object!']}
_RESP_MISSING_ARGS = {'code': '5102', 'message': ['Required argument field(s) missing...']}
_RESP_SIGN_CHECK_FAIL = {'code': '5104', 'message': ['Invalid sign, sign check failed!']}


class APIHandler(web.RequestHandler):
    LOG = LOG

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abc.abstractmethod
    def response(self, *args, **kwargs) -> dict:
        raise NotImplementedError()

    def set_default_headers(self) -> None:
        self.set_header('Content-Type', 'application/json; charset=utf-8')

    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    async def post(self):
        try:
            req = self.request.body
            body = json.loads(req.decode('utf-8'))
        except json.decoder.JSONDecodeError:  # invalid request body, cannot be parsed as JSON
            return self.finish(_RESP_BAD_REQUEST)

        try:
            salt_uuid = body.pop('salt_uuid')
            app_id = body.pop('app_id')
            sign = body.pop('sign')
            data = body.pop('data')
        except KeyError:  # cannot find default key from parsed body
            return self.finish(_RESP_MISSING_ARGS)

        is_valid_req = sign_check(salt_uuid=salt_uuid, app_id=app_id, sign=sign, data=data)  # , sign_method='sha256'
        if not is_valid_req:
            return self.finish(_RESP_SIGN_CHECK_FAIL)

        resp = dict(code=5200, message=['success'])
        try:
            result = self.response(**data)  # this call may throw TypeError when argument missing
            resp['data'] = result
            resp['salt_uuid'] = salt_uuid
        except Exception as e:
            if LOG.level == logging.DEBUG:
                self.LOG.error(e, exc_info=True)
            return self.finish({'code': 5201, 'message': [str(e)]})

        resp = json.dumps(resp, ensure_ascii=False, default=str, separators=(',', ':'))
        return self.finish(resp)



from secrets import SystemRandom

random = SystemRandom()


def random_bool():
    return random.choice([True, False])


def random_choice(*args, **kwargs):
    return random.choice(*args, **kwargs)


def random_int(a, b):
    return random.randint(a, b)


def random_ratio():
    return random.random()


def random_uniform(*args):
    return random.uniform(*args)


def random_sample(*args):
    return random.sample(*args)


def random_seed(*args, **kwargs):
    return random.seed(*args, **kwargs)



__all__ = ('Application',)

import asyncio

from ..logger import LOG

try:
    import uvloop
    from tornado.platform.asyncio import AsyncIOMainLoop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    AsyncIOMainLoop().install()
    # asyncio.get_event_loop().run_forever()
    # ^ the line above should be replaced by the start function below.
    LOG.info('Using uvloop for service event loop...')
except ImportError:
    LOG.warn('[uvloop] NOT installed, fallback to asyncio loop! Please `pip install uvloop`!')

from .web import WebApplication
from ..settings import SETTINGS

from tornado.options import options

options['log_file_prefix'] = 'access.log'

io_loop = asyncio.get_event_loop()


class Application:
    def __init__(self, *args, **kwargs):
        settings = dict(SETTINGS.config)
        self.web_app = WebApplication(settings)
        self.io_loop = io_loop

    def start(self):
        try:
            self.web_app.start()
            self.io_loop.run_forever()
        except KeyboardInterrupt:
            pass
        except Exception as e:
            raise e
        finally:
            pass

    def stop(self):
        self.io_loop.stop()



__all__ = ('RedisOperator',)

import redis

from .base import password_vault
from ..logger import LOG


class RedisOperator:
    def __init__(self, config):
        self._check_redis_version()

        password = config.get('password', None)
        _config = {
            'host': config['host'],
            'port': config.get('port', '6379'),
            'password': password_vault.get_password(password),
            'decode_responses': config.get('decode_responses', True),
            'retry_on_timeout': True,
            'max_connections': config.get('max_connections', 1000),
            'socket_timeout': 3,
            'socket_connect_timeout': 1,
        }
        if 'db' in config:
            _config['db'] = config['db']
        self._config = _config

        self._pool = None

    @staticmethod
    def _check_redis_version() -> bool:
        valid = False
        try:
            ver = redis.__version__.split('.')
            ver = tuple(int(i) for i in ver)
            if ver > (4, 1, 0):
                valid = True
                LOG.debug('Using redis version = %s' % redis.__version__)
        except Exception as e:
            LOG.error('Failed to obtain redis version!')
            LOG.error(str(e))

        if not valid:
            msg = 'Invalid version of `redis-py`, version >4.1.0 required!'
            LOG.fatal(msg)
            raise ImportError(msg)

        return valid

    @property
    def connection_generic(self):
        """https://github.com/redis/redis-py/blob/master/redis/client.py"""
        LOG.debug("StrictRedis connection info: {host}:{port}".format(**self._config))

        if self._pool is None:
            self._pool = redis.ConnectionPool()
        return redis.Redis(connection_pool=self._pool, **self._config)

    @property
    def connection_cluster(self):
        LOG.debug("RedisCluster connection info: {host}:{port}".format(**self._config))
        return redis.RedisCluster(**self._config)



import argparse
import importlib


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('cmd')
    args, _ = parser.parse_known_args()

    cmd = args.cmd
    module = '%s.%s' % (__package__, cmd)
    try:
        module = importlib.import_module(module)
    except ImportError:
        print('Invalid sub-command: %s\n\tFailed to import: %s' % (cmd, module))
        exit(-1)

    func_main = getattr(module, 'main')

    exit(func_main())


if __name__ == '__main__':
    main()



__all__ = ('ElasticSearchOperator',)

import json

from elasticsearch import Elasticsearch

from .base import password_vault
from ..logger import LOG


class ElasticSearchOperator:
    def __init__(self, config, index_config=None):
        self.es_config = config

        username = config.get('username')
        password = password_vault.get_password(config.get('password'))

        self._config = {
            'http_auth': (username, password) if username is not None and password is not None else None,
            'hosts': config.get('host', 'localhost'),

            'timeout': config.get("timeout", 0.1),
            'max_retries': config.get("max_retries", 3),
            'retry_on_timeout': config.get("retry_on_timeout", True)
        }
        LOG.debug("ElasticSearch connection info: " + str(self._config['hosts']))

        self.index_config = index_config
        self.index_name = self.es_config.get('index_name')
        self.index_type = self.es_config.get('index_type')

        self.es = Elasticsearch(**self._config)

        if index_config is not None:
            self.index_config = self._load_config(index_config)

    @staticmethod
    def _load_config(config):
        if type(config) == dict:
            return config

        elif type(config) == str and ".json" in config:
            with open(config, "r", encoding="utf-8") as f:
                config = json.load(f)
            return config

        else:
            raise Exception("Invalid ES config data type")

    def put_mapping(self, index_name=None, index_type=None, index_config=None):
        return self.es.indices.put_mapping(
            index=index_name or self.index_name,
            doc_type=index_type or self.index_type,
            body=index_config["mappings"][index_type or self.index_type]
        )

    def build_index(self, index_name=None, index_config=None, raise_if_exist=False):
        if self.es.indices.exists(index=index_name or self.index_name) is not True:
            res = self.es.indices.create(
                index=index_name or self.index_name,
                body=index_config or self.index_config
            )
            return res
        else:
            msg = "Index [%s] already exits" % self.index_name
            if raise_if_exist:
                raise RuntimeError(msg)
            else:
                LOG.info(msg)
                return False

    def search(self, query, index_name=None, index_type=None):
        return self.es.search(index=index_name or self.index_name, doc_type=index_type or self.index_type, body=query)

    def msearch(self, body):
        return self.es.msearch(body=body)

    def insert(self, doc, index_name=None, index_type=None, id=None):
        return self.es.index(index=index_name or self.index_name, doc_type=index_type or self.index_type, id=id, body=doc)



import hashlib
import json


def get_md5_of_str(string):
    return hashlib.md5(string.encode()).hexdigest()


def get_sha256_of_str(string):
    return hashlib.sha256(string.encode()).hexdigest()


def hash_dict(dic):
    s = json.dumps(dict(dic), sort_keys=True, ensure_ascii=False, default=str)
    return hashlib.md5(s.encode()).hexdigest()


def hash_obj(obj):
    s = json.dumps(obj, sort_keys=True, ensure_ascii=False, default=str)
    return hashlib.md5(s.encode()).hexdigest()



import logging
import os
import socket
from os.path import join as pjoin

from .handler import MultiProcessSafeDailyRotatingFileHandler


def setup_logger(logger: logging.Logger, level: int = logging.DEBUG, logger_name: str = None, module: str = None, formatter_str: str = None):
    if not logger.handlers:
        formatter = logging.Formatter(formatter_str or '%(levelname)s> %(asctime)s> %(module)s:%(lineno)s> %(message)s')

        folder = 'logs'
        os.makedirs(folder, exist_ok=True)

        if module is None:
            from ..settings import SETTINGS
            module = SETTINGS.config.get('APP_MODULE') or os.environ.get('APP_MODULE', 'default')

        if logger_name is not None and len(logger_name) > 0:
            module = '%s_%s' % (logger_name, module)

        path_file = [module, socket.gethostname(), 'p%s' % os.getpid()]  # module, hostname, pid
        path_file = '_'.join(str(i) for i in path_file if i is not None)
        path_file = pjoin(folder, '%s.log' % path_file)
        file_handler = MultiProcessSafeDailyRotatingFileHandler(path_file)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)

        logger.setLevel(level)


def get_logger(level=logging.DEBUG, logger_name: str = None, *args, **kwargs) -> logging.Logger:
    logger = logging.getLogger(logger_name)

    if isinstance(level, str):
        level = getattr(logging, str(level).upper(), 10)

    setup_logger(logger, level=level, logger_name=logger_name, *args, **kwargs)
    return logger



# ! `pip install pynvml`, reference: https://github.com/gpuopenanalytics/pynvml

from collections import namedtuple

from ..logger import LOG

try:
    import pynvml
except ImportError:
    raise RuntimeError('Please install package `pynvml` first!')

Device = namedtuple('Device', field_names='index,name,arch')
DeviceStatus = namedtuple('DeviceStatus', field_names='mem_total,mem_free,mem_used,gpu_rate,mem_rate')


class NvInfo:
    def __init__(self):
        self.nv = pynvml
        self.nv.nvmlInit()
        ver: bytes = self.nv.nvmlSystemGetDriverVersion()
        version = ver.decode(encoding='UTF-8')
        LOG.info('Nvidia Driver version: %s' % version)

    def __del__(self):
        self.nv.nvmlShutdown()

    @property
    def get_device_count(self) -> int:
        try:
            return self.nv.nvmlDeviceGetCount()
        except pynvml.NVMLError as e:
            return 0

    def get_device_list(self) -> list:
        list_device = []
        for i in range(self.get_device_count):
            handler = self._get_device_handler(index_device=i)
            print(self.nv.nvmlDeviceGetArchitecture(handler))
            device = Device(
                index=i,
                name=self.nv.nvmlDeviceGetName(handler).decode(encoding='UTF-8'),
                arch=self.nv.nvmlDeviceGetArchitecture(handler)
            )

            list_device.append(device)

        return list_device

    def _get_device_handler(self, index_device=0):
        return self.nv.nvmlDeviceGetHandleByIndex(index_device)

    def get_device_status(self, index_device=0) -> DeviceStatus:
        handler = self._get_device_handler(index_device=index_device)
        mem_info = self.nv.nvmlDeviceGetMemoryInfo(handler)
        util_info = self.nv.nvmlDeviceGetUtilizationRates(handler)
        return DeviceStatus(
            mem_total=mem_info.total,
            mem_free=mem_info.free,
            mem_used=mem_info.used,
            gpu_rate=util_info.gpu,
            mem_rate=util_info.memory
        )


def main(*args, **kwargs):
    import json
    n = NvInfo()
    list_devices = n.get_device_list()
    for i, d in enumerate(list_devices):
        status = n.get_device_status(index_device=i)
        item = {
            'device': d._asdict(),
            'status': status._asdict()
        }
        print(json.dumps(item))


if __name__ == '__main__':
    main()



from .base import BaseVault
from ...encrypt.aes import AesEncryptor


class PlainVault(AesEncryptor, BaseVault):
    def __init__(self, key: str = None):
        super().__init__(key)

    def decrypt_password(self, password):
        return self.decrypt(password)


def main():
    vault = PlainVault()
    pwd = vault.get_password(None, url_quote=True)
    # print(pwd)
    return pwd



import jwt

from ..logger import LOG

LOG.debug('Version of pyjwt = %s' % jwt.__version__.__str__())


def encode(
        secret_key: str,
        payload: dict,
        headers: dict = None,
        **kwargs
):
    token = jwt.encode(payload=payload, key=secret_key, headers=headers, **kwargs)
    return token


def decode(
        secret_key: str,
        token: str,
        **kwargs
):
    try:
        resp = jwt.decode(jwt=token, key=secret_key, algorithms=['HS256'], **kwargs)
    except jwt.ExpiredSignatureError as e:
        resp = str(e)
    except jwt.PyJWTError as e:
        resp = str(e)
    return resp



import asyncio
import enum
import sys
import warnings
from functools import wraps, partial
from types import TracebackType
from typing import Any, Optional, Type

# from typing_extensions import final


__all__ = ("timeout", "timeout_at")


def timeout(delay: Optional[float]) -> "Timeout":
    """timeout context manager.
    Useful in cases when you want to apply timeout logic around block
    of code or in cases when asyncio.wait_for is not suitable. For example:
    >>> async with timeout(0.001):
    ...     async with aiohttp.get('https://github.com') as r:
    ...         await r.text()
    delay - value in seconds or None to disable timeout logic
    """
    loop = _get_running_loop()
    if delay is not None:
        deadline = loop.time() + delay  # type: Optional[float]
    else:
        deadline = None
    return Timeout(deadline, loop)


def timeout_at(deadline: Optional[float]) -> "Timeout":
    """Schedule the timeout at absolute time.
    deadline arguments points on the time in the same clock system
    as loop.time().
    Please note: it is not POSIX time but a time with
    undefined starting base, e.g. the time of the system power on.
    >>> async with timeout_at(loop.time() + 10):
    ...     async with aiohttp.get('https://github.com') as r:
    ...         await r.text()
    """
    loop = _get_running_loop()
    return Timeout(deadline, loop)


class _State(enum.Enum):
    INIT = "INIT"
    ENTER = "ENTER"
    TIMEOUT = "TIMEOUT"
    EXIT = "EXIT"


# @final
class Timeout:
    # Internal class, please don't instantiate it directly
    # Use timeout() and timeout_at() public factories instead.
    #
    # Implementation note: `async with timeout()` is preferred
    # over `with timeout()`.
    # While technically the Timeout class implementation
    # doesn't need to be async at all,
    # the `async with` statement explicitly points that
    # the context manager should be used from async function context.
    #
    # This design allows to avoid many silly misusages.
    #
    # TimeoutError is raised immadiatelly when scheduled
    # if the deadline is passed.
    # The purpose is to time out as sson as possible
    # without waiting for the next await expression.

    __slots__ = ("_deadline", "_loop", "_state", "_task", "_timeout_handler")

    def __init__(
            self, deadline: Optional[float], loop: asyncio.AbstractEventLoop
    ) -> None:
        self._loop = loop
        self._state = _State.INIT

        task = _current_task(self._loop)
        self._task = task

        self._timeout_handler = None  # type: Optional[asyncio.Handle]
        if deadline is None:
            self._deadline = None  # type: Optional[float]
        else:
            self.shift_to(deadline)

    def __enter__(self) -> "Timeout":
        warnings.warn(
            "with timeout() is deprecated, use async with timeout() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        self._do_enter()
        return self

    def __exit__(
            self,
            exc_type: Type[BaseException],
            exc_val: BaseException,
            exc_tb: TracebackType,
    ) -> Optional[bool]:
        self._do_exit(exc_type)
        return None

    async def __aenter__(self) -> "Timeout":
        self._do_enter()
        return self

    async def __aexit__(
            self,
            exc_type: Type[BaseException],
            exc_val: BaseException,
            exc_tb: TracebackType,
    ) -> Optional[bool]:
        self._do_exit(exc_type)
        return None

    @property
    def expired(self) -> bool:
        """Is timeout expired during execution?"""
        return self._state == _State.TIMEOUT

    @property
    def deadline(self) -> Optional[float]:
        return self._deadline

    def reject(self) -> None:
        """Reject scheduled timeout if any."""
        # cancel is maybe better name but
        # task.cancel() raises CancelledError in asyncio world.
        if self._state not in (_State.INIT, _State.ENTER):
            raise RuntimeError("invalid state {}".format(self._state.value))
        self._reject()

    def _reject(self) -> None:
        if self._timeout_handler is not None:
            self._timeout_handler.cancel()
            self._timeout_handler = None

    def shift_by(self, delay: float) -> None:
        """Advance timeout on delay seconds.
        The delay can be negative.
        """
        now = self._loop.time()
        self.shift_to(now + delay)

    def shift_to(self, deadline: float) -> None:
        """Advance timeout on the abdelay seconds.
        If new deadline is in the past
        the timeout is raised immediatelly.
        """
        if self._state == _State.EXIT:
            raise RuntimeError("cannot reschedule after exit from context manager")
        if self._state == _State.TIMEOUT:
            raise RuntimeError("cannot reschedule expired timeout")
        if self._timeout_handler is not None:
            self._timeout_handler.cancel()
        self._deadline = deadline
        now = self._loop.time()
        if deadline <= now:
            self._timeout_handler = None
            if self._state == _State.INIT:
                raise asyncio.TimeoutError
            else:
                # state is ENTER
                raise asyncio.CancelledError
        self._timeout_handler = self._loop.call_at(
            deadline, self._on_timeout, self._task
        )

    def _do_enter(self) -> None:
        if self._state != _State.INIT:
            raise RuntimeError("invalid state {}".format(self._state.value))
        self._state = _State.ENTER

    def _do_exit(self, exc_type: Type[BaseException]) -> None:
        if exc_type is asyncio.CancelledError and self._state == _State.TIMEOUT:
            self._timeout_handler = None
            raise asyncio.TimeoutError
        # timeout is not expired
        self._state = _State.EXIT
        self._reject()
        return None

    def _on_timeout(self, task: "asyncio.Task[None]") -> None:
        task.cancel()
        self._state = _State.TIMEOUT


def _current_task(loop: asyncio.AbstractEventLoop) -> "Optional[asyncio.Task[Any]]":
    if sys.version_info >= (3, 7):
        return asyncio.current_task(loop=loop)
    else:
        return asyncio.Task.current_task(loop=loop)


def _get_running_loop() -> asyncio.AbstractEventLoop:
    loop = asyncio.get_running_loop()
    if loop is None:
        print('--' * 20)
        loop = asyncio.get_event_loop()

    if sys.version_info >= (3, 7):
        return loop
    else:
        if not loop.is_running():
            raise RuntimeError("no running event loop")
        return loop


def aioify(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        p_func = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, p_func)

    return run



from tornado import web
from typing import Optional, Awaitable


class BaseHandler(web.RequestHandler):
    def data_received(self, chunk: bytes) -> Optional[Awaitable[None]]:
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_default_headers(self):
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Access-Control-Max-Age', 1000)
        self.set_header('Content-type', 'application/json; charset=UTF-8')
        self.set_header('Access-Control-Allow-Methods', 'POST, GET, OPTIONS')
        self.set_header(
            'Access-Control-Allow-Headers',
            'authorization, Authorization, Content-Type,'
            'Access-Control-Allow-Origin, Access-Control-Allow-Headers,'
            'X-Requested-By, Access-Control-Allow-Methods'
        )



from binascii import b2a_hex, a2b_hex

import pyhocon
from Crypto.Cipher import AES


def _is_empty_str(s):
    return s is None or isinstance(s, pyhocon.config_tree.NoneValue) or s == "None" or s == ""


class AesEncryptor:
    def __init__(self, key: str = None):
        self.key = key or 'b6046801716aec00'

    def auto_fill_key(self):
        __key = self.key
        if len(__key) > 32:
            raise ValueError("Key length should be less than 32!")
        while len(__key) not in [16, 24, 32]:
            __key += " "
        return __key.encode()

    def encrypt(self, text):
        if _is_empty_str(text):
            return None

        text = text.encode('utf-8')
        cryptor = AES.new(self.auto_fill_key(), AES.MODE_CBC, b'0000000000000000')
        length = 16
        count = len(text)
        if count < length:
            add = (length - count)
            text = text + (b'\0' * add)
        elif count > length:
            add = (length - (count % length))
            text = text + (b'\0' * add)
        ciphertext = cryptor.encrypt(text)
        return b2a_hex(ciphertext).decode()

    def decrypt(self, text):
        if _is_empty_str(text):
            return None

        cryptor = AES.new(self.auto_fill_key(), AES.MODE_CBC, b'0000000000000000')
        plain_text = cryptor.decrypt(a2b_hex(text.encode()))
        return plain_text.rstrip(b'\0').decode()
